model: add read-acquire/fence-release support
[c11tester.git] / action.cc
index 1387ed15b67c9ac25d7fcba68491bbb72d8ff236..a90878f73d9edc34148359f127df6886d4574cd8 100644 (file)
--- a/action.cc
+++ b/action.cc
@@ -31,13 +31,14 @@ ModelAction::ModelAction(action_type_t type, memory_order order, void *loc,
        location(loc),
        value(value),
        reads_from(NULL),
+       last_fence_release(NULL),
        node(NULL),
        seq_number(ACTION_INITIAL_CLOCK),
        cv(NULL),
        sleep_flag(false)
 {
        /* References to NULL atomic variables can end up here */
-       ASSERT(loc || type == MODEL_FIXUP_RELSEQ);
+       ASSERT(loc || type == ATOMIC_FENCE || type == MODEL_FIXUP_RELSEQ);
 
        Thread *t = thread ? thread : thread_current();
        this->tid = t->get_id();
@@ -69,6 +70,11 @@ void ModelAction::set_seq_number(modelclock_t num)
        seq_number = num;
 }
 
+bool ModelAction::is_thread_start() const
+{
+       return type == THREAD_START;
+}
+
 bool ModelAction::is_relseq_fixup() const
 {
        return type == MODEL_FIXUP_RELSEQ;
@@ -156,6 +162,11 @@ bool ModelAction::is_initialization() const
        return type == ATOMIC_INIT;
 }
 
+bool ModelAction::is_relaxed() const
+{
+       return order == std::memory_order_relaxed;
+}
+
 bool ModelAction::is_acquire() const
 {
        switch (order) {
@@ -248,13 +259,15 @@ bool ModelAction::could_synchronize_with(const ModelAction *act) const
        if (!same_var(act))
                return false;
 
-       // Explore interleavings of seqcst writes to guarantee total order
-       // of seq_cst operations that don't commute
-       if ((could_be_write() || act->could_be_write()) && is_seqcst() && act->is_seqcst())
+       // Explore interleavings of seqcst writes/fences to guarantee total
+       // order of seq_cst operations that don't commute
+       if ((could_be_write() || act->could_be_write() || is_fence() || act->is_fence())
+                       && is_seqcst() && act->is_seqcst())
                return true;
 
-       // Explore synchronizing read/write pairs
-       if (is_read() && is_acquire() && act->could_be_write() && act->is_release())
+       // Explore synchronizing read/write/fence pairs
+       if (is_acquire() && act->is_release() && (is_read() || is_fence()) &&
+                       (act->could_be_write() || act->is_fence()))
                return true;
 
        //lock just released...we can grab lock
@@ -329,24 +342,10 @@ void ModelAction::set_try_lock(bool obtainedlock) {
 /**
  * Update the model action's read_from action
  * @param act The action to read from; should be a write
- * @return True if this read established synchronization
  */
-bool ModelAction::read_from(const ModelAction *act)
+void ModelAction::set_read_from(const ModelAction *act)
 {
-       ASSERT(cv);
        reads_from = act;
-       if (act != NULL && this->is_acquire()) {
-               rel_heads_list_t release_heads;
-               model->get_release_seq_heads(this, &release_heads);
-               int num_heads = release_heads.size();
-               for (unsigned int i = 0; i < release_heads.size(); i++)
-                       if (!synchronize_with(release_heads[i])) {
-                               model->set_bad_synchronization();
-                               num_heads--;
-                       }
-               return num_heads > 0;
-       }
-       return false;
 }
 
 /**
@@ -468,22 +467,22 @@ void ModelAction::print() const
                break;
        }
 
-       printf("(%4d) Thread: %-2d   Action: %-13s   MO: %7s  Loc: %14p   Value: %-#18" PRIx64,
+       model_print("(%4d) Thread: %-2d   Action: %-13s   MO: %7s  Loc: %14p   Value: %-#18" PRIx64,
                        seq_number, id_to_int(tid), type_str, mo_str, location, valuetoprint);
        if (is_read()) {
                if (reads_from)
-                       printf("  Rf: %-3d", reads_from->get_seq_number());
+                       model_print("  Rf: %-3d", reads_from->get_seq_number());
                else
-                       printf("  Rf: ?  ");
+                       model_print("  Rf: ?  ");
        }
        if (cv) {
                if (is_read())
-                       printf(" ");
+                       model_print(" ");
                else
-                       printf("          ");
+                       model_print("          ");
                cv->print();
        } else
-               printf("\n");
+               model_print("\n");
 }
 
 /** @brief Print nicely-formatted info about this ModelAction */